#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
Module: brokers.virtual.virtual_broker
Purpose: AutoTrader virtual broker for backtesting
Author: Kieran Mackle
'''

import numpy as np

class Broker():
    """
    Virtual broker to simulate trading environment.


    Attributes
    ----------
    broker_config : dict
        the broker configuration dictionary
    
    leverage : int
        the leverage on the account (default 1)
    
    commission : int
        the brokers commission as a percentage (default 0)
        
    spread : int
        the bid/ask spread (default 0)
    
    utils : class
        the broker utilities class
    
    home_curreny : str
        the home currency of the account (used for retrieving quote data)


    Methods
    -------
    place_order(order_details):
        Places an order with the virtual broker.
    
    open_position(order_no, candle, limit_price = None):
        Opens an order to enter the market as a trade. 
    
    update_positions(candle):
        Updates orders and open positions based on current candle. 

    
    
    """
    
    def __init__(self, broker_config, utils):
        self.leverage           = 1
        self.commission         = 0
        self.spread             = 0
        self.margin_available   = 0
        self.portfolio_balance  = 0
        self.pending_orders  = {}
        self.open_positions     = {}
        self.closed_positions   = {}
        self.cancelled_orders   = {}
        self.total_trades       = 0
        self.profitable_trades  = 0
        self.peak_value         = 0
        self.low_value          = 0
        self.max_drawdown       = 0
        self.home_currency      = 'AUD'
        self.NAV                = 0
        self.unrealised_PL      = 0
        self.utils              = utils
        self.verbosity          = broker_config['verbosity']
        

    def place_order(self, order_details):
        '''
            Place order with broker.
        '''
        instrument  = order_details["instrument"]
        size        = order_details["size"]
        stop_loss  = order_details["stop_loss"]
        stop_distance = order_details["stop_distance"]
        take_profit = order_details["take_profit"]
        
        if order_details['order_type'] == 'market':
            order_price = order_details["order_price"]
        elif order_details['order_type'] == 'limit' or order_details['order_type'] == 'stop-limit':
            order_price = order_details["order_limit_price"]
        else:
            order_price = order_details["order_price"]
        
        if stop_loss is None and stop_distance is not None:
            pip_value   = self.utils.get_pip_ratio(instrument)
            stop_loss  = order_price - np.sign(size)*stop_distance*pip_value
        
        order_no = self.total_trades + 1
        new_position = order_details.copy()
        new_position['order_ID'] = order_no
        
        # Verify SL and TP prices
        invalid_order = False
        if stop_loss is not None and np.sign(size)*(order_price - stop_loss) < 0:
            direction = 'long' if np.sign(size) > 0 else 'short'
            SL_placement = 'below' if np.sign(size) > 0 else 'above'
            if self.verbosity > 0:
                print("Invalid stop loss request: stop loss must be "+ \
                                f"{SL_placement} the order price for a {direction}" + \
                                " trade order.\n"+ \
                                f"Order Price: {order_price}\nStop Loss: {stop_loss}")
            invalid_order = True
        
        if take_profit is not None and np.sign(size)*(order_price - take_profit) > 0:
            direction = 'long' if np.sign(size) > 0 else 'short'
            TP_placement = 'above' if np.sign(size) > 0 else 'below'
            if self.verbosity > 0:
                print("Invalid take profit request: take profit must be "+ \
                                f"{TP_placement} the order price for a {direction}" + \
                                " trade order.\n"+ \
                                f"Order Price: {order_price}\nTake Profit: {take_profit}")
            invalid_order = True
        
        if invalid_order:
            if self.verbosity > 0:
                print(f"  Order {order_no} rejected.\n")
            # Add position to self.pending_orders
            self.cancelled_orders[order_no] = new_position
            self.cancelled_orders[order_no]['reason'] = "Invalid order request"
            
        else:
            # Add position to self.pending_orders
            self.pending_orders[order_no] = new_position
            
        # Update trade tally
        self.total_trades += 1
    
    
    def open_position(self, order_no, candle, limit_price = None):
        ''' Opens position with broker. '''
        
        # Calculate margin requirements
        current_price   = candle.Open
        pip_value       = self.utils.get_pip_ratio(self.pending_orders[order_no]['instrument'])
        size            = self.pending_orders[order_no]['size']
        HCF             = self.pending_orders[order_no]['HCF']
        position_value  = abs(size) * current_price * HCF
        margin_required = self.calculate_margin(position_value)
        
        spread_cost = abs(size) * self.spread * pip_value
        spread_shift = 0.5 * np.sign(size) * self.spread * pip_value
        
        if margin_required < self.margin_available:
            # Fill order
            new_position = self.pending_orders[order_no]
            new_position['time_filled'] = candle.name
            if limit_price is None:
                new_position['entry_price'] = candle.Open + spread_shift
            else:
                new_position['entry_price'] = limit_price + spread_shift
            self.open_positions[order_no] = new_position
            
            # Subtract spread cost
            self.NAV -= spread_cost
            
        else:
            # Cancel order
            cancel_reason = "Insufficient margin"
            self.cancel_pending_order(order_no, cancel_reason)
    
    def update_positions(self, candle, instrument):
        ''' 
            Updates orders and open positions based on current candle. 
        '''

        # Tally for positions opened this update
        opened_positions = 0
        modification_orders = []
        
        # Update pending positions
        for order_no in self.pending_orders:
            # Filter orders by instrument type since candle is instrument specific            
            if self.pending_orders[order_no]['instrument'] == instrument:
                if self.pending_orders[order_no]['order_time'] != candle.name:
                    if self.pending_orders[order_no]['order_type'] == 'market':
                        # Market order type
                        self.open_position(order_no, candle)
                        opened_positions += 1
                    
                    elif self.pending_orders[order_no]['order_type'] == 'stop-limit':
                        # Stop-limit order type
                        # Check if order_stop_price has been reached yet
                        
                        if candle.Low < self.pending_orders[order_no]['order_stop_price'] < candle.High:
                            # order_stop_price has been reached, change order type to 'limit'
                            self.pending_orders[order_no]['order_type'] = 'limit'
                    
                    elif self.pending_orders[order_no]['order_type'] == 'modify':
                        # Modification order
                        self.modify_order(self.pending_orders[order_no])
                        
                        opened_positions += 1 # To remove from pending orders
                        modification_orders.append(order_no)
                        
                    # This is in a separate if statement, as stop-limit order may
                    # eventually be changed to limit orders
                    if self.pending_orders[order_no]['order_type'] == 'limit':
                        # Limit order type
                        if self.pending_orders[order_no]['size'] > 0:
                            if candle.Low < self.pending_orders[order_no]['order_limit_price']:
                                self.open_position(order_no, candle, 
                                                   self.pending_orders[order_no]['order_limit_price'])
                                opened_positions += 1
                        else:
                            if candle.High > self.pending_orders[order_no]['order_limit_price']:
                                self.open_position(order_no, candle, 
                                                   self.pending_orders[order_no]['order_limit_price'])
                                opened_positions += 1
                                
                                
                if self.pending_orders[order_no]['order_type'] == 'close':
                    related_order = self.pending_orders[order_no]['related_orders']
                    self.close_position(self.pending_orders[order_no]['instrument'],
                                        candle, 
                                        candle.Close,
                                        order_no = related_order
                                        )
                    opened_positions += 1 # To remove from pending orders
                    modification_orders.append(order_no)
                elif self.pending_orders[order_no]['order_type'] == 'reduce':
                    self.reduce_position(self.pending_orders[order_no])
        
                
        # Remove position from pending positions
        if opened_positions > 0:
            # For orders that were opened
            for order_no in self.open_positions.keys():
                self.pending_orders.pop(order_no, 0)
            
            # For orders that were cancelled
            for order_no in self.cancelled_orders.keys():
                self.pending_orders.pop(order_no, 0)
            
            # For close orders
            for order_no in modification_orders:
                self.pending_orders.pop(order_no, 0)
        
        
        # Update trailing stops
        # For other methods, move the stop update to an external function
        # Can this be moved into the loop below?
        for order_no in self.open_positions:
            if self.open_positions[order_no]['instrument'] == instrument:
                if self.open_positions[order_no]['stop_type'] == 'trailing':
                    # Trailing stop loss is enabled, check if price has moved SL
                    
                    if self.open_positions[order_no]['stop_distance'] is not None:
                        # Stop distance provided 
                        pip_value = self.utils.get_pip_ratio(self.open_positions[order_no]['instrument'])
                        pip_distance = self.open_positions[order_no]['stop_distance']
                        distance = pip_distance*pip_value
                        
                    else:
                        # Stop loss price provided
                        distance = abs(self.open_positions[order_no]['entry_price'] - \
                                       self.open_positions[order_no]['stop_loss'])
                        
                        # Append stop distance to dict
                        pip_value = self.utils.get_pip_ratio(self.open_positions[order_no]['instrument'])
                        self.open_positions[order_no]['stop_distance'] = distance / pip_value
                        
    
                    if self.open_positions[order_no]['size'] > 0:
                        # long position, stop loss only moves up
                        new_stop = candle.High - distance
                        if new_stop > self.open_positions[order_no]['stop_loss']:
                            self._update_stop_loss(order_no, new_stop)
                        
                    else:
                        # short position, stop loss only moves down
                        new_stop = candle.Low + distance
                        if new_stop < self.open_positions[order_no]['stop_loss']:
                            self._update_stop_loss(order_no, new_stop)
        
        # Update self.open_positions
        open_position_orders = list(self.open_positions.keys())
        unrealised_PL  = 0        # Un-leveraged value
        for order_no in open_position_orders:
            if self.open_positions[order_no]['instrument'] == instrument:
                if self.open_positions[order_no]['size'] > 0:
                    # Long trade
                    if self.open_positions[order_no]['stop_loss'] is not None and \
                        candle.Low < self.open_positions[order_no]['stop_loss']:
                        # Stop loss hit
                        self.close_position(self.open_positions[order_no]['instrument'], 
                                            candle, 
                                            self.open_positions[order_no]['stop_loss'],
                                            order_no)
                    elif self.open_positions[order_no]['take_profit'] is not None and \
                        candle.High > self.open_positions[order_no]['take_profit']:
                        # Take Profit hit
                        self.close_position(self.open_positions[order_no]['instrument'], 
                                            candle, 
                                            self.open_positions[order_no]['take_profit'],
                                            order_no)
                    else:
                        # Position is still open, update value of holding
                        size        = self.open_positions[order_no]['size']
                        entry_price = self.open_positions[order_no]['entry_price']
                        price       = candle.Close
                        HCF         = self.open_positions[order_no]['HCF']
                        trade_PL    = size*(price - entry_price)*HCF
                        unrealised_PL += trade_PL
                        
                        # Update PL of trade
                        self.open_positions[order_no]['last_price'] = price
                        self.open_positions[order_no]['last_time'] = candle.name
                        self.open_positions[order_no]['unrealised_PL'] = trade_PL
                
                else:
                    # Short trade
                    if self.open_positions[order_no]['stop_loss'] is not None and \
                        candle.High > self.open_positions[order_no]['stop_loss']:
                        # Stop loss hit
                        self.close_position(self.open_positions[order_no]['instrument'], 
                                            candle, 
                                            self.open_positions[order_no]['stop_loss'],
                                            order_no)
                    elif self.open_positions[order_no]['take_profit'] is not None and \
                        candle.Low < self.open_positions[order_no]['take_profit']:
                        # Take Profit hit
                        self.close_position(self.open_positions[order_no]['instrument'], 
                                            candle, 
                                            self.open_positions[order_no]['take_profit'],
                                            order_no)
                    else:
                        # Position is still open, update value of holding
                        size        = self.open_positions[order_no]['size']
                        entry_price = self.open_positions[order_no]['entry_price']
                        price       = candle.Close
                        HCF         = self.open_positions[order_no]['HCF']
                        trade_PL    = size*(price - entry_price)*HCF
                        unrealised_PL += trade_PL
                        
                        # Update PL of trade
                        self.open_positions[order_no]['last_price'] = price
                        self.open_positions[order_no]['last_time'] = candle.name
                        self.open_positions[order_no]['unrealised_PL'] = trade_PL
        
        # Update margin available
        self.update_margin()
        
        # Update unrealised P/L
        self.unrealised_PL = unrealised_PL
        
        # Update open position value
        self.NAV = self.portfolio_balance + self.unrealised_PL
    
    
    def close_position(self, instrument, candle, exit_price,
                           order_no=None):
        ''' Closes position. '''
        if order_no is not None:
            # single order specified to close
            self.close_trade(order_no=order_no, 
                             candle=candle, 
                             exit_price=exit_price)
        else:
            # Close all positions for instrument
            for order_no in self.open_positions.copy():
                if self.open_positions[order_no]['instrument'] == instrument:
                    self.close_trade(order_no=order_no, 
                                     candle=candle, 
                                     exit_price=exit_price)
    
    
    def close_trade(self, candle=None, exit_price=None,
                           order_no=None):
        ''' Closes trade by order number. '''
        entry_price = self.open_positions[order_no]['entry_price']
        size        = self.open_positions[order_no]['size']
        HCF         = self.open_positions[order_no]['HCF']
        
        # Update portfolio with profit/loss
        gross_PL    = size*(exit_price - entry_price)*HCF
        open_trade_value    = abs(size)*entry_price*HCF
        close_trade_value   = abs(size)*exit_price*HCF
        commission  = (self.commission/100) * (open_trade_value + close_trade_value)
        net_profit  = gross_PL - commission
        
        self.add_funds(net_profit)
        if net_profit > 0:
            self.profitable_trades += 1
        
        # Add trade to closed positions
        closed_position = self.open_positions[order_no].copy()
        closed_position['profit'] = net_profit
        closed_position['balance'] = self.portfolio_balance
        closed_position['exit_price'] = exit_price
        closed_position['fees'] = commission
        if candle is None:
            closed_position['exit_time'] = self.open_positions[order_no]['last_time']
        else:
            closed_position['exit_time'] = candle.name 
        
        # Add to closed positions dictionary
        self.closed_positions[order_no] = closed_position
        
        # Remove position from open positions
        self.open_positions.pop(order_no, 0)
        
        # Update maximum drawdown
        self.update_MDD()
    
    def reduce_position(self, order_details):
        ''' 
        Reduces the position of the specified instrument by FIFO. 
        
        The size parameter of the order details is used to specify whether 
        to reduce long units, or to reduce short units. For example, the
        order details below will reduce long units of the position being 
        held.
            order_type: 'reduce' # reduction order
            size: -1 # reduce long units by selling
        '''
        
        # Consired long vs. short units to be reduced
        instrument = order_details['instrument']
        reduction_size = order_details['size']
        reduction_direction = order_details['direction']
        
        # Get open trades for instrument
        open_trades = self.get_open_trades(instrument)
        open_trade_IDs = list(open_trades.keys())
        
        # Modify existing trades until there are no more units to reduce
        units_to_reduce = reduction_size
        while units_to_reduce > 0:
            for order_no in open_trade_IDs:
                if open_trades[order_no]['direction'] != reduction_direction:
                    # Only reduce long trades when reduction direction is -1
                    # Only reduce short trades when reduction direction is 1
                    if units_to_reduce > open_trades[order_no]['size']:
                        # Entire trade must be closed
                        last_price = open_trades[order_no]['last_price']
                        self.close_trade(order_no = order_no, exit_price=last_price)
                        
                        # Update units_to_reduce
                        units_to_reduce -= abs(self.closed_positions[order_no]['size'])
                        
                    else:
                        # Partially close trade
                        self.partial_trade_close(order_no, units_to_reduce)
                        
                        # Update units_to_reduce
                        units_to_reduce = 0
                    
    
    def partial_trade_close(self, trade_ID, units):
        ''' Partially closes a trade. '''
        entry_price = self.open_positions[trade_ID]['entry_price']
        size        = self.open_positions[trade_ID]['size']
        HCF         = self.open_positions[trade_ID]['HCF']
        last_price  = self.open_positions[trade_ID]['last_price']
        remaining_size = size-units
        
        # Update portfolio with profit/loss
        gross_PL    = units*(last_price - entry_price)*HCF
        open_trade_value    = abs(units)*entry_price*HCF
        close_trade_value   = abs(units)*last_price*HCF
        commission  = (self.commission/100) * (open_trade_value + close_trade_value)
        net_profit  = gross_PL - commission
        
        self.add_funds(net_profit)
        
        if net_profit > 0:
            self.profitable_trades += 1
        
        # Add trade to closed positions
        closed_position = self.open_positions[trade_ID].copy()
        closed_position['size'] = units
        closed_position['profit'] = net_profit
        closed_position['balance'] = self.portfolio_balance
        closed_position['exit_price'] = last_price
        closed_position['exit_time'] = self.open_positions[trade_ID]['last_time']
        closed_position['fees'] = commission
        
        # Add to closed positions dictionary
        partial_trade_close_ID = self.total_trades + 1
        self.closed_positions[partial_trade_close_ID] = closed_position
        self.total_trades += 1
        
        # Modify remaining portion of trade
        self.open_positions[trade_ID]['size'] = remaining_size
        
        # Update maximum drawdown
        self.update_MDD()
        
    
    def get_pending_orders(self, instrument = None):
        ''' Returns pending orders. '''
        
        pending_orders = {}
        
        if instrument is not None:
            for order_no in self.pending_orders:
                if self.pending_orders[order_no]['instrument'] == instrument:
                    pending_orders[order_no] = self.pending_orders[order_no]
        else:
            pending_orders = self.pending_orders.copy()
        
        return pending_orders
    
    def cancel_pending_order(self, order_id, reason):
        ''' Moves an order from pending_orders into cancelled_orders. '''
        
        self.cancelled_orders[order_id] = self.pending_orders[order_id]
        self.cancelled_orders[order_id]['reason'] = reason
    
    def get_open_trades(self, instruments=None):
        ''' Returns open trades for the specified instrument. '''
        
        # Check data type of input
        if instruments is not None:
            # Non-None input provided, check type
            if type(instruments) is str:
                # Single instrument provided, put into list
                instruments = [instruments]
                
        open_trades = {}
        
        if instruments is not None:
            # Specific instruments requested
            for order_no in self.open_positions:
                if self.open_positions[order_no]['instrument'] in instruments:
                    open_trades[order_no] = self.open_positions[order_no]
        else:
            # Return all currently open positions
            open_trades = self.open_positions.copy()
        
        return open_trades
    
    def get_trade_details(self, trade_ID):
        'Returns the details of the trade specified by trade_ID.'
        return self.open_positions[trade_ID]
    
    def get_open_positions(self, instruments=None):
        ''' Returns the open positions in the account. '''
        
        if instruments is None:
            # No specific instrument requested, get all open
            instruments = []
            for order_no in self.open_positions: 
                instruments.append(self.open_positions[order_no]['instrument'])
        else:
            # Non-None input provided, check type
            if type(instruments) is str:
                # Single instrument provided, put into list
                instruments = [instruments]
        
        open_positions = {}
        for instrument in instruments:
            # First get open trades
            open_trades = self.get_open_trades(instrument)
            
            if len(open_trades) > 0:
                # Trades exist for current instrument, collate
                long_units = 0
                long_PL = 0
                long_margin = 0
                short_units = 0
                short_PL = 0
                short_margin = 0
                total_margin = 0
                trade_IDs = []
                
                for order_no in open_trades:
                    trade_IDs.append(order_no)
                    total_margin += open_trades[order_no]['margin_required']
                    if open_trades[order_no]['size'] > 0:
                        # Long trade
                        long_units += open_trades[order_no]['size']
                        long_PL += open_trades[order_no]['unrealised_PL']
                        long_margin += open_trades[order_no]['margin_required']
                    else:
                        # Short trade
                        short_units += open_trades[order_no]['size']
                        short_PL += open_trades[order_no]['unrealised_PL']
                        short_margin += open_trades[order_no]['margin_required']
            
                # Construct instrument position dict
                instrument_position = {'long_units': long_units,
                                        'long_PL': long_PL,
                                        'long_margin': long_margin,
                                        'short_units': short_units,
                                        'short_PL': short_PL,
                                        'short_margin': short_margin,
                                        'total_margin': total_margin,
                                        'trade_IDs': trade_IDs}
                
                # Append position dict to open_positions dict
                open_positions[instrument] = instrument_position
                        
        return open_positions
    
    def get_cancelled_orders(self, instrument = None):
        ''' Returns cancelled orders. '''
        
        cancelled_orders = {}
        
        if instrument is not None:
            for order_no in self.cancelled_orders:
                if self.cancelled_orders[order_no]['instrument'] == instrument:
                    cancelled_orders[order_no] = self.cancelled_orders[order_no]
        else:
            cancelled_orders = self.cancelled_orders.copy()
        
        return cancelled_orders
    
    
    def add_funds(self, amount):
        ''' Add's funds to brokerage account. '''
        
        self.portfolio_balance  += amount
    
    def make_deposit(self, deposit):
        '''
        Adds deposit to account balance and NAV.
        '''
        
        if self.portfolio_balance == 0:
            # If this is the initial deposit, set peak and low values for MDD
            self.peak_value = deposit
            self.low_value = deposit
            
        self.portfolio_balance += deposit
        self.NAV += deposit
        self.update_margin()
    
    def get_balance(self):
        ''' Returns balance of account. '''
        return self.portfolio_balance
    
    
    def calculate_margin(self, position_value):
        ''' Calculates margin required to take a position. '''
        margin = position_value / self.leverage
        
        return margin
    
    
    def update_margin(self):
        ''' Updates margin available in account. '''
        
        margin_used = 0
        for order_no in self.open_positions:
            size            = self.open_positions[order_no]['size']
            HCF             = self.open_positions[order_no]['HCF']
            last_price      = self.open_positions[order_no]['last_price']
            position_value  = abs(size) * last_price * HCF
            margin_required = self.calculate_margin(position_value)
            margin_used     += margin_required
            
            # Update margin required in trade dict
            self.open_positions[order_no]['margin_required'] = margin_required
        
        self.margin_available = self.portfolio_balance - margin_used


    def get_price(self, instrument, data=None, conversion_data=None, i=None):
        ''' Returns the price data dict. '''
        
        if data is not None and conversion_data is not None and i is not None:
            ask = data.Close[i]
            bid = data.Close[i]
            conversion_data = conversion_data.Close[i]
            
            if bid == conversion_data:
                negativeHCF = 1
                positiveHCF = 1
            else:
                negativeHCF = 1/conversion_data
                positiveHCF = 1/conversion_data
        else:
            # Allow calling get_price as placeholder for livetrading
            ask = data.Close[i]
            bid = data.Close[i]
            negativeHCF = 1
            positiveHCF = 1
        
        price = {"ask": ask,
                 "bid": bid,
                 "negativeHCF": negativeHCF,
                 "positiveHCF": positiveHCF
                 }
        
        return price
    
    def update_MDD(self):
        ''' Function to calculate maximum portfolio drawdown '''
        balance     = self.portfolio_balance
        peak_value  = self.peak_value
        low_value   = self.low_value
        
        if balance > peak_value:
            self.peak_value = balance
            self.low_value = balance
            
        elif balance < low_value:
            self.low_value = balance
        
        MDD = 100*(low_value - peak_value)/peak_value
        
        if MDD < self.max_drawdown:
            self.max_drawdown = MDD
    
    def get_NAV(self):
        ''' Returns Net Asset Value of account. '''
        return self.NAV
    
    def get_margin_available(self):
        ''' Returns the margin available on the account. '''
        return self.margin_available
    
    def modify_order(self, modification_order):
        ''' 
        Modify order with updated parameters. Called when order_type = 'modify', 
        modifies trade specified by related_orders key.
        '''
        
        # Get ID of trade to modify
        modify_trade_id = modification_order['related_orders']
        
        if modification_order['stop_loss'] is not None:
            # New stop loss provided
            self._update_stop_loss(modify_trade_id, 
                                   modification_order['stop_loss'],
                                   modification_order['stop_type'])
            
        if modification_order['take_profit'] is not None:
            self._update_take_profit(modify_trade_id, 
                                     modification_order['take_profit'])
        
    def _update_stop_loss(self, trade_id, new_stop_loss, new_stop_type='limit'):
        ''' Updates stop loss on open trade. '''
        self.open_positions[trade_id]['stop_loss'] = new_stop_loss
        self.open_positions[trade_id]['stop_type'] = new_stop_type
        self.open_positions[trade_id]['stop_distance'] = None
    
    def _update_take_profit(self, trade_id, new_take_profit):
        ''' Updates take profit on open trade. '''
        self.open_positions[trade_id]['take_profit'] = new_take_profit


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
Module: brokers.binanceapi.utils
Purpose: Utility functions for Binance broker wrapper
Author: Kieran Mackle
'''

import pandas as pd
from datetime import datetime
from autotrader.brokers.broker_utils import BrokerUtils


class Utils(BrokerUtils):
    def __init__(self):
        return
    
    def bars_to_df(self, bars):
        ''' Function to convert Binance klines to Pandas dataframe. '''
        for line in bars:
            del line[5:]
            
        data = pd.DataFrame(bars, columns=['date', 'Open', 'High', 'Low', 'Close'])
        dates = [datetime.fromtimestamp(timestamp/1000) for timestamp in data.date.values]
        
        data.index = dates
        
        return data.drop(['date'], axis=1)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
Module: lib.indicators
Purpose: AutoTrader custom indicators library
Author: Kieran Mackle
'''

from finta import TA
import numpy as np
import pandas as pd

''' -------------------------- PRICE INDICATORS --------------------------- '''
def supertrend(data, period = 10, ATR_multiplier = 3.0, source=None):
    ''' Based on the SuperTrend indicator by KivancOzbilgic on TradingView '''
    
    if source is None:
        source = (data.High.values + data.Low.values) / 2
    
    # Calculate ATR
    atr             = TA.ATR(data, period)
    
    up              = source - (ATR_multiplier*atr)
    up_list         = [up[0]]
    up_times        = [data.index[0]]
    N_up            = 0
    
    dn              = source + (ATR_multiplier*atr)
    dn_list         = [dn[0]]
    dn_times        = [data.index[0]]
    N_dn            = 0
    
    trend           = 1
    trend_list      = [trend]
    
    for i in range(1, len(data)):
        
        if trend == 1:
            if data.Close.values[i] > max(up[N_up:i]):
                up_list.append(max(up[N_up:i]))
                up_times.append(data.index[i])
                
                dn_list.append(np.nan)
                dn_times.append(data.index[i])
                
            else: 
                trend = -1
                N_dn = i
                dn_list.append(dn[i])
                dn_times.append(data.index[i])
                
                up_list.append(np.nan)
                up_times.append(data.index[i])
                
        else:
            if data.Close.values[i] < min(dn[N_dn:i]):
                dn_list.append(min(dn[N_dn:i]))
                dn_times.append(data.index[i])
                
                up_list.append(np.nan)
                up_times.append(data.index[i])
                
            else:
                trend = 1
                N_up = i
                up_list.append(up[i])
                up_times.append(data.index[i])
                
                dn_list.append(np.nan)
                dn_times.append(data.index[i])
        
        trend_list.append(trend)
    
    # up_trend = pd.DataFrame({'uptrend': up_list}, index = up_times)
    # dn_trend = pd.DataFrame({'downtrend': dn_list}, index = dn_times)
    
    supertrend_df = pd.DataFrame({'uptrend': up_list,
                                  'downtrend': dn_list,
                                  'trend': trend_list}, 
                                 index = up_times)
    
    return supertrend_df


def stoch_rsi(data, K_period=3, D_period=3, RSI_length=14, Stochastic_length=14):

    rsi1 = TA.RSI(data, period=RSI_length)
    stoch = stochastic(rsi1, rsi1, rsi1, Stochastic_length)
    
    K = sma(stoch, K_period)
    D = sma(K, D_period)
    
    return K, D


def stochastic(high, low, close, period=14):
    
    K = np.zeros(len(high))
    
    for i in range(period, len(high)):
        low_val     = min(low[i-period+1:i+1])
        high_val    = max(high[i-period+1:i+1])
        
        K[i]        = 100 * (close[i] - low_val)/(high_val - low_val)
        
    return K
    

def sma(data, period=14):
    
    sma_list = []
    
    for i in range(len(data)):
        average = sum(data[i-period+1:i+1])/period
        sma_list.append(average)
    
    return sma_list


def ema(data, period=14, smoothing=2):
    
    ema = [sum(data[:period]) / period]
    
    for price in data[period:]:
        ema.append((price * (smoothing / (1 + period))) + ema[-1] * (1 - (smoothing / (1 + period))))
    
    for i in range(period-1):
        ema.insert(0, np.nan)
    
    return ema


def true_range(data, period=14):
    high_low = data['High'] - data['Low']
    high_close = np.abs(data['High'] - data['Close'].shift())
    low_close = np.abs(data['Low'] - data['Close'].shift())
    
    ranges = pd.concat([high_low, high_close, low_close], axis=1)
    true_range = np.max(ranges, axis=1)
    
    return true_range


def atr(data, period=14):
    
    tr = true_range(data, period)
    
    atr = tr.rolling(period).sum()/period
    
    return atr


def bullish_engulfing(data, detection = None):
    ''' Bullish engulfing pattern detection. '''
    
    if detection == "SMA50":
        sma50       = sma(data.Close.values, 50)
        
        down_trend  = np.where(data.Close.values < sma50, True, False)
        up_trend    = np.where(data.Close.values > sma50, True, False)
        
    elif detection == "SMA50/200":
        sma50       = sma(data.Close.values, 50)
        sma200      = sma(data.Close.values, 200)
        
        down_trend  = np.where((data.Close.values < sma50) & 
                               (data.Close.values < sma200), 
                               True, False)
        up_trend    = np.where((data.Close.values > sma50) & 
                               (data.Close.values > sma200), 
                               True, False)
    else:
        down_trend  = np.full(len(data), True)
        up_trend    = np.full(len(data), True)
    
    body_len        = 14    # ema depth for bodyAvg
    shadow_pc       = 100.0   # size of shadows 
    doji_pc         = 5.0
    shadow_factor   = 2.0 # number of times shadow dominates candle body
    
    body_high       = np.maximum(data.Close.values, data.Open.values)
    body_low        = np.minimum(data.Close.values, data.Open.values)
    body            = body_high - body_low
    
    body_avg        = ema(body, body_len)
    short_body      = body < body_avg
    long_body       = body > body_avg
    up_shadow       = data.High.values - body_high
    down_shadow     = body_low - data.Low.values
    
    has_up_shadow   = up_shadow > (shadow_pc / 100 * body)
    has_dn_shadow   = down_shadow > (shadow_pc / 100 * body)
    white_body      = data.Open.values < data.Close.values
    black_body      = data.Open.values > data.Close.values
    candle_range    = data.High.values - data.Low.values
    
    inside_bar = [False]
    for i in range(1, len(data)):
        val  = (body_high[i-1] > body_high[i]) and (body_low[i-1] < body_low[i])
        inside_bar.append(val)
        
    body_mid        = body/2 + body_low
    shadow_equals   = (up_shadow == down_shadow) | \
                          (
                              ((abs(up_shadow - down_shadow) / down_shadow * 100) < shadow_pc) & \
                              ((abs(down_shadow - up_shadow) / up_shadow * 100) < shadow_pc)
                          )
    doji_body       = (candle_range > 0) & (body <= candle_range * doji_pc / 100)
    doji            = doji_body & shadow_equals
    
    
    engulfing_bullish = [False]
    for i in range(1, len(data)):
        condition = down_trend[i] & \
                    white_body[i] & \
                    long_body[i] & \
                    black_body[i-1] & \
                    short_body[i-1] & \
                    (data.Close.values[i] >= data.Open.values[i-1]) & \
                    (data.Open.values[i] <= data.Close.values[i-1]) & \
                    ((data.Close.values[i] > data.Open.values[i-1]) | (data.Open.values[i] < data.Close.values[i-1]))
        
        engulfing_bullish.append(condition)
        
    return engulfing_bullish


def bearish_engulfing(data, detection = None):
    ''' Bearish engulfing pattern detection. '''
    
    if detection == "SMA50":
        sma50       = sma(data.Close.values, 50)
        
        down_trend  = np.where(data.Close.values < sma50, True, False)
        up_trend    = np.where(data.Close.values > sma50, True, False)
        
    elif detection == "SMA50/200":
        sma50       = sma(data.Close.values, 50)
        sma200      = sma(data.Close.values, 200)
        
        down_trend  = np.where((data.Close.values < sma50) & 
                               (data.Close.values < sma200), 
                               True, False)
        up_trend    = np.where((data.Close.values > sma50) & 
                               (data.Close.values > sma200), 
                               True, False)
    else:
        down_trend  = np.full(len(data), True)
        up_trend    = np.full(len(data), True)
    
    body_len        = 14    # ema depth for bodyAvg
    shadow_pc       = 100.0   # size of shadows 
    doji_pc         = 5.0
    shadow_factor   = 2.0 # number of times shadow dominates candle body
    
    body_high       = np.maximum(data.Close.values, data.Open.values)
    body_low        = np.minimum(data.Close.values, data.Open.values)
    body            = body_high - body_low
    
    body_avg        = ema(body, body_len)
    short_body      = body < body_avg
    long_body       = body > body_avg
    up_shadow       = data.High.values - body_high
    down_shadow     = body_low - data.Low.values
    
    has_up_shadow   = up_shadow > (shadow_pc / 100 * body)
    has_dn_shadow   = down_shadow > (shadow_pc / 100 * body)
    white_body      = data.Open.values < data.Close.values
    black_body      = data.Open.values > data.Close.values
    candle_range    = data.High.values - data.Low.values
    
    inside_bar = [False]
    for i in range(1, len(data)):
        val  = (body_high[i-1] > body_high[i]) and (body_low[i-1] < body_low[i])
        inside_bar.append(val)
        
    body_mid        = body/2 + body_low
    shadow_equals   = (up_shadow == down_shadow) | \
                          (
                              ((abs(up_shadow - down_shadow) / down_shadow * 100) < shadow_pc) & \
                              ((abs(down_shadow - up_shadow) / up_shadow * 100) < shadow_pc)
                          )
    doji_body       = (candle_range > 0) & (body <= candle_range * doji_pc / 100)
    doji            = doji_body & shadow_equals
    
    
    engulfing_bearish = [False]
    for i in range(1, len(data)):
        condition = up_trend[i] & \
                    black_body[i] & \
                    long_body[i] & \
                    white_body[i-1] & \
                    short_body[i-1] & \
                    (data.Close.values[i] <= data.Open.values[i-1]) & \
                    (data.Open.values[i] >= data.Close.values[i-1]) & \
                    ((data.Close.values[i] < data.Open.values[i-1]) | (data.Open.values[i] > data.Close.values[i-1]))
        
        engulfing_bearish.append(condition)
        
    return engulfing_bearish


def heikin_ashi(data):
    ''' 
        Calculates the Heikin-Ashi candlesticks from Japanese candlestick 
        data. 
    '''
    
    # Create copy of data to prevent overwriting
    working_data = data.copy()
    
    # Calculate Heikin Ashi candlesticks
    ha_close = 0.25*(working_data.Open + working_data.Low + working_data.High + working_data.Close)
    ha_open = 0.5*(working_data.Open + working_data.Close)
    ha_high = np.maximum(working_data.High.values, working_data.Close.values, working_data.Open.values)
    ha_low = np.minimum(working_data.Low.values, working_data.Close.values, working_data.Open.values)
    
    ha_data = pd.DataFrame(data={'Open': ha_open, 
                                 'High': ha_high, 
                                 'Low': ha_low, 
                                 'Close': ha_close}, 
                           index=working_data.index)
    
    return ha_data
 
def half_trend(data, amplitude=2, channel_deviation=2):
    '''
    HalfTrend indicator, originally by Alex Orekhov (everget) on TradingView.
    
    Parameters:
        data (dataframe): OHLC price data
        
        amplitude (int): lookback window
            
        channel_deviation (int): ATR channel deviation factor
    
    (translated to Python 15/11/21)
    '''
    
    # Initialisation
    atr2 = TA.ATR(data, 100)/2
    dev = channel_deviation * atr2
    high_price = data.High.rolling(amplitude).max().fillna(0)
    low_price = data.Low.rolling(amplitude).min().fillna(0)
    highma = TA.SMA(data, period=amplitude, column='High')
    lowma = TA.SMA(data, period=amplitude, column='Low')
    
    trend = np.zeros(len(data))
    next_trend = np.zeros(len(data))
    max_low_price = np.zeros(len(data))
    max_low_price[0] = data.Low[0]
    min_high_price = np.zeros(len(data))
    min_high_price[0] = data.High[0]
    
    for i in range(1, len(data)):
        if next_trend[i-1] == 1:
            max_low_price[i] = max(low_price[i-1], max_low_price[i-1])

            if highma[i] < max_low_price[i] and data.Close[i] < data.Low[i-1]:
                trend[i] = 1
                next_trend[i] = 0
                min_high_price[i] = high_price[i]
            else:
                # assign previous values again
                trend[i] = trend[i-1]
                next_trend[i] = next_trend[i-1]
                min_high_price[i] = min_high_price[i-1]
        else:
            min_high_price[i] = min(high_price[i-1], min_high_price[i-1])
            
            if lowma[i] > min_high_price[i] and data.Close[i] > data.High[i-1]:
                trend[i] = 0
                next_trend[i] = 1
                max_low_price[i] = low_price[i]
            else:
                # assign previous values again
                trend[i] = trend[i-1]
                next_trend[i] = next_trend[i-1]
                max_low_price[i] = max_low_price[i-1]
                
    up = np.zeros(len(data))
    up[0] = max_low_price[0]
    down = np.zeros(len(data))
    down[0] = min_high_price[0]
    atr_high = np.zeros(len(data))
    atr_low = np.zeros(len(data))
    
    for i in range(1, len(data)):
        if trend[i] == 0:
            if trend[i-1] != 0:
                up[i] = down[i-1]
            else:
                up[i] = max(max_low_price[i-1], up[i-1])
            
            atr_high[i] = up[i] + dev[i]
            atr_low[i] = up[i] - dev[i]
            
        else:
            if trend[i-1] != 1:
                down[i] = up[i-1]
            else:
                down[i] = min(min_high_price[i-1], down[i-1]) 
                
            atr_high[i] = down[i] + dev[i]
            atr_low[i] = down[i] - dev[i]
    
    halftrend = np.where(trend == 0, up, down)
    buy = np.where((trend==0) & (np.roll(trend,1)==1), 1, 0)
    sell = np.where((trend==1) & (np.roll(trend,1)==0), 1, 0)
    
    # Construct DataFrame
    htdf = pd.DataFrame(data = {'halftrend': halftrend, 
                                'atrHigh': np.nan_to_num(atr_high),
                                'atrLow': np.nan_to_num(atr_low),
                                'buy': buy, 
                                'sell': sell},
                        index = data.index)
    
    # Clear false leading signals
    htdf.buy.values[:100] = np.zeros(100)
    htdf.sell.values[:100] = np.zeros(100)
    
    # Replace leading zeroes with nan
    htdf['atrHigh'] = htdf.atrHigh.replace(to_replace=0, value=float("nan"))
    htdf['atrLow'] = htdf.atrLow.replace(to_replace=0, value=float("nan"))
    
    return htdf

def N_period_high(data, N):
    ''' Returns the N-period high. '''
    highs = data.High.rolling(N).max()
    return highs

def N_period_low(data, N):
    ''' Returns the N-period low. '''
    lows = data.Low.rolling(N).min()
    return lows

''' ------------------------ UTILITY INDICATORS --------------------------- '''

def crossover(list_1, list_2):
    ''' 
    Returns a list of length len(list_1) with 1 when list_1 crosses above
    list_2 and -1 when list_1 crosses below list_2.
    '''
    
    sign_list = []
    for i in range(len(list_1)):
        if np.isnan(list_1[i]):
            sign_list.append(np.nan)
        else:
            difference = list_1[i] - list_2[i]
            if difference < 0:
                sign_list.append(-1)
            else:
                sign_list.append(1)
    
    crossover_list = [0]
    
    for i in range(1, len(sign_list)):
        if sign_list[i] - sign_list[i-1] != 0:
            val = sign_list[i]
        else:
            val = 0
        
        crossover_list.append(val)

    return crossover_list


def cross_values(a, b, ab_crossover):
    cross_point_list = [0]
    last_cross_point = 0
    for i in range(1, len(ab_crossover)):
        if ab_crossover[i] != 0:
            i0 = 0
            m_a = a[i] - a[i-1]
            m_b = b[i] - b[i-1]
            ix = (b[i-1] - a[i-1])/(m_a-m_b) + i0
            
            cross_point = m_a*(ix - i0) + a[i-1]
            
            last_cross_point = cross_point
            
        else:
            cross_point = last_cross_point #0
        
        cross_point_list.append(cross_point)
    
    # Replace nans with 0
    cross_point_list = [0 if x!=x  else x for x in cross_point_list]
    
    return cross_point_list


def candles_between_crosses(cross_list):
    '''
    Returns candles since last cross
    
    
    Behaviour:
    in:  [0, 0, 1, 0, 0, 0, -1, 0, 0, 0, 0, 0, 1]
    out: [1, 2, 0, 1, 2, 3,  0, 1, 2, 3, 4, 5, 0]
    '''
    
    count = 0
    count_list = []
    
    for i in range(len(cross_list)):

        if cross_list[i] == 0:
            # Change in signal - reset count
            count += 1
        else:
            count = 0
        
        count_list.append(count)
    
    return count_list

def find_swings(data, data_type='ohlc', n = 2):
    '''
    Locates swings in the inputted data and returns a dataframe.
    
    Parameters:
        data: an OHLC dataframe of price, or an array/list of data from an 
        indicator.
        
        data_type: specify 'ohlc' when data is OHLC, or 'other' when inputting
        an indicator.
    '''
    
    # Prepare data 
    if data_type == 'ohlc':
        # Find swings in OHLC price data
        # hl2     = (data.Open.values + data.Close.values)/2
        hl2     = (data.High.values + data.Low.values)/2
        swing_data = ema(hl2, n)
        
        low_data = data.Low.values
        high_data = data.High.values
        
    else:
        # Find swings in alternative data source
        swing_data = data
        low_data = data
        high_data = data
    
    # Calculate slope of data and points where slope changes
    grad = [swing_data[i] - swing_data[i-1] for i in range(len(swing_data))]
    swings = np.where(np.sign(grad) != np.sign(np.roll(grad,1)), -np.sign(grad), 0)

    # Construct columns
    low_list    = [0,0]
    high_list   = [0,0]
    for i in range(2,len(data)):
        if swings[i] == -1:
            # Down swing - find min price in the vicinity 
            high_list.append(0)
            low_list.append(min(low_data[i-n:i]))
            
        elif swings[i] == 1:
            # Up swing - find max price in the vicinity
            high_list.append(max(high_data[i-n:i]))
            low_list.append(0)
            
        else:
            # Price movement
            low_list.append(0)
            high_list.append(0)
    
    trend = rolling_signal_list(-swings)
    swings_list = merge_signals(low_list, high_list)
    last_swing = rolling_signal_list(swings_list)
    last_swing[0:n] = list(high_data[0:n])
    
    # Need to return both a last swing low and last swing high list
    last_low = rolling_signal_list(low_list)
    last_low[0:n] = list(low_data[0:n])     # Fill start of data
    last_high = rolling_signal_list(high_list)
    last_high[0:n] = list(high_data[0:n])   # Fill start of data
    
    swing_df = pd.DataFrame(data={'Highs': last_high, 
                                  'Lows' : last_low,
                                  'Last' : last_swing,
                                  'Trend': trend},
                            index = data.index)
    
    return swing_df
    
def classify_swings(swing_df, tol=0):
    ''' 
    Classify a dataframe of swings (from find_swings) into higher-high, 
    lower-high, higher-low and lower-low.
    
    Parameters:
        swing_df: the dataframe outputted from find_swings.
        
        tol: parameter to control strength of levels detected.
    '''
    
    # Create copy of swing dataframe
    swing_df = swing_df.copy()
    
    new_level = np.where(swing_df.Last != swing_df.Last.shift(), 1, 0)

    candles_since_last = candles_between_crosses(new_level)
    
    # Add column 'candles since last swing' CSLS
    swing_df['CSLS'] = candles_since_last
    
    # Find strong Support and Resistance zones
    swing_df['Support'] = (swing_df.CSLS > tol) & (swing_df.Trend == 1)
    swing_df['Resistance'] = (swing_df.CSLS > tol) & (swing_df.Trend == -1)
    
    # Find higher highs and lower lows
    swing_df['Strong_lows'] = swing_df['Support'] * swing_df['Lows'] # Returns high values when there is a strong support
    swing_df['Strong_highs'] = swing_df['Resistance'] * swing_df['Highs'] # Returns high values when there is a strong support
    
    # Remove duplicates to preserve indexes of new levels
    swing_df['FSL'] = unroll_signal_list(swing_df['Strong_lows']) # First of new strong lows
    swing_df['FSH'] = unroll_signal_list(swing_df['Strong_highs']) # First of new strong highs
    
    # Now compare each non-zero value to the previous non-zero value.
    low_change = np.sign(swing_df.FSL) * (swing_df.FSL - swing_df.Strong_lows.replace(to_replace=0, method='ffill').shift())
    high_change = np.sign(swing_df.FSH) * (swing_df.FSH - swing_df.Strong_highs.replace(to_replace=0, method='ffill').shift())
    
    swing_df['LL'] = np.where(low_change < 0, True, False)
    swing_df['HL'] = np.where(low_change > 0, True, False)
    swing_df['HH'] = np.where(high_change > 0, True, False)
    swing_df['LH'] = np.where(high_change < 0, True, False)
    
    return swing_df

def detect_divergence(classified_price_swings, classified_indicator_swings, tol=2,
                      method=0):
    '''
    Detects divergence between price swings and swings in an indicator.
    
    Parameters:
        classified_price_swings: output from classify_swings using OHLC data.
        
        classified_indicator_swings: output from classify_swings using indicator data.
        
        tol: number of candles which conditions must be met within. 
        
        method: the method to use when detecting divergence. Options include:
            0: use both price and indicator swings to detect divergence (default)
            1: use only indicator swings to detect divergence
    '''
    
    if method == 0:
        
        regular_bullish = []
        regular_bearish = []
        hidden_bullish = []
        hidden_bearish = []
        
        for i in range(len(classified_price_swings)):
            # Look backwards in each
            
            # REGULAR BULLISH DIVERGENCE
            if sum(classified_price_swings['LL'][i-tol:i]) + sum(classified_indicator_swings['HL'][i-tol:i]) > 1:
                regular_bullish.append(True)
            else:
                regular_bullish.append(False)
            
            # REGULAR BEARISH DIVERGENCE
            if sum(classified_price_swings['HH'][i-tol:i]) + sum(classified_indicator_swings['LH'][i-tol:i]) > 1:
                regular_bearish.append(True)
            else:
                regular_bearish.append(False)
            
            # HIDDEN BULLISH DIVERGENCE
            if sum(classified_price_swings['HL'][i-tol:i]) + sum(classified_indicator_swings['LL'][i-tol:i]) > 1:
                hidden_bullish.append(True)
            else:
                hidden_bullish.append(False)
            
            # HIDDEN BEARISH DIVERGENCE
            if sum(classified_price_swings['LH'][i-tol:i]) + sum(classified_indicator_swings['HH'][i-tol:i]) > 1:
                hidden_bearish.append(True)
            else:
                hidden_bearish.append(False)
            
            divergence = pd.DataFrame(data = {'regularBull': unroll_signal_list(regular_bullish),
                                              'regularBear': unroll_signal_list(regular_bearish),
                                              'hiddenBull': unroll_signal_list(hidden_bullish),
                                              'hiddenBear': unroll_signal_list(hidden_bearish)})
    elif method == 1:
        # Use indicator swings only to detect divergence
        for i in range(len(classified_price_swings)):
            
            price_at_indi_lows = (classified_indicator_swings['FSL'] != 0) * classified_price_swings['Lows']
            price_at_indi_highs = (classified_indicator_swings['FSH'] != 0) * classified_price_swings['Highs']
            
            # Determine change in price between indicator lows
            price_at_indi_lows_change = np.sign(price_at_indi_lows) * (price_at_indi_lows - price_at_indi_lows.replace(to_replace=0, method='ffill').shift())
            price_at_indi_highs_change = np.sign(price_at_indi_highs) * (price_at_indi_highs - price_at_indi_highs.replace(to_replace=0, method='ffill').shift())
            
            
            # DETECT DIVERGENCES
            regular_bullish = (classified_indicator_swings['HL']) & (price_at_indi_lows_change < 0 )
            regular_bearish = (classified_indicator_swings['LH']) & (price_at_indi_highs_change > 0 )
            hidden_bullish = (classified_indicator_swings['LL']) & (price_at_indi_lows_change > 0 )
            hidden_bearish = (classified_indicator_swings['HH']) & (price_at_indi_highs_change < 0 )
            
            divergence = pd.DataFrame(data = {'regularBull': regular_bullish,
                                              'regularBear': regular_bearish,
                                              'hiddenBull': hidden_bullish,
                                              'hiddenBear': hidden_bearish})
        
    else:
        raise Exception("Error: unrecognised method of divergence detection.")
        
    return divergence

def autodetect_divergence(ohlc, indicator_data, method=0):
    '''
    Wrapper method to automatically detect divergence from inputted OHLC price 
    data and indicator data.
    
    This method calls:
        find_swings()
        classify_swings()
        detect_divergence()
    
    Parameters:
        ohlc: dataframe of OHLC data
        
        indicator data: array of indicator data
        
        method: the method to use when detecting divergence. Options include:
            0: use both price and indicator swings to detect divergence (default)
            1: use only indicator swings to detect divergence
    '''
    
    # Price swings
    price_swings = find_swings(ohlc)
    price_swings_classified = classify_swings(price_swings)
    
    # Indicator swings 
    indicator_swings = find_swings(indicator_data, data_type='other')
    indicator_classified = classify_swings(indicator_swings)
    
    # Detect divergence
    divergence = detect_divergence(price_swings_classified, indicator_classified, method)
    
    return divergence

def rolling_signal_list(signals):
    ''' 
        Returns a list which maintains the previous signal, until a new 
        signal is given.
        
        [0,1,0,0,0,-1,0,0,1,0,0] ->  [0,1,1,1,1,-1,-1,-1,1,1,1]
        
    '''
    
    rolling_signals = [0]
    last_signal     = rolling_signals[0]
    
    for i in range(1, len(signals)):
        if signals[i] != 0:
            last_signal = signals[i]
        
        rolling_signals.append(last_signal)
    
    return rolling_signals

def unroll_signal_list(signals):
    ''' Unrolls a signal list. '''
    new_list = np.zeros(len(signals))
    
    for i in range(len(signals)):
        if signals[i] != signals[i-1]:
            new_list[i] = signals[i]
    
    return new_list


def merge_signals(signal_1, signal_2):
     ''' 
         Returns a single signal list which has merged two signal lists. 
     '''
     
     merged_signal_list = signal_1.copy()
     
     for i in range(len(signal_1)):
         if signal_2[i] != 0:
             merged_signal_list[i] = signal_2[i]
     
     return merged_signal_list


def ha_candle_run(ha_data):
    '''
        Returns a list for the number of consecutive green and red 
        Heikin-Ashi candles.
        
    '''
    green_candle    = np.where(ha_data.Close - ha_data.Open > 0, 1, 0)
    red_candle      = np.where(ha_data.Close - ha_data.Open < 0, 1, 0)
    
    green_run   = []
    red_run     = []
    
    green_sum   = 0
    red_sum     = 0
    
    for i in range(len(ha_data)):
        if green_candle[i] == 1:
            green_sum += 1
        else:
            green_sum = 0
        
        if red_candle[i] == 1:
            red_sum += 1
        else:
            red_sum = 0
        
        green_run.append(green_sum)
        red_run.append(red_sum)
        
    return green_run, red_run


def build_grid_price_levels(grid_origin, grid_space, grid_levels, 
                            grid_price_space=None, pip_value=0.0001):
    
    # Calculate grid spacing in price units
    if grid_price_space is None:
        grid_price_space = grid_space*pip_value
    
    # Generate order_limit_price list 
    grid_price_levels = np.linspace(grid_origin - grid_levels*grid_price_space, 
                                     grid_origin + grid_levels*grid_price_space, 
                                     2*grid_levels + 1)
    
    return grid_price_levels

def build_grid(grid_origin, grid_space, grid_levels, order_direction, 
               order_type='stop-limit', grid_price_space=None, pip_value=0.0001, 
               take_distance=None, stop_distance=None, stop_type=None):
    '''
    grid_origin: origin of grid, specified as a price
    grid_space: spacing between grid levels, specified as pip distance
    grid_levels: number of grid levels either side of origin
    order_direction: the direction of each grid level order (1 for long, -1 for short)
    order_type: the order type of each grid level order
    '''
    # TODO - could add a limit price buffer, then use it to move the limit price
    # slightly away from the stop price
    
    # Check if stop_distance was provided without a stop_type
    if stop_distance is not None and stop_type is None:
        # set stop_type to 'limit' by default
        stop_type = 'limit'
    
    # Calculate grid spacing in price units
    if grid_price_space is None:
        grid_price_space = grid_space*pip_value
    
    # Generate order_limit_price list 
    order_limit_prices = np.linspace(grid_origin - grid_levels*grid_price_space, 
                                     grid_origin + grid_levels*grid_price_space, 
                                     2*grid_levels + 1)
    
    # Construct nominal order
    nominal_order = {}
    nominal_order["order_type"]         = order_type
    nominal_order["direction"]          = order_direction
    nominal_order["stop_distance"]      = stop_distance
    nominal_order["stop_type"]          = stop_type
    nominal_order["take_distance"]      = take_distance
    
    # Build grid
    grid = {}

    for order, limit_price in enumerate(order_limit_prices):
        grid[order] = nominal_order.copy()
        grid[order]["order_stop_price"]  = order_limit_prices[order]
        grid[order]["order_limit_price"] = order_limit_prices[order]
        
    
    return grid

def merge_grid_orders(grid_1, grid_2):
    '''
    Merges grid dictionaries into one and re-labels order numbers so each
    order number is unique.
    '''
    # TODO - use **args/**kwargs to generalise how many grids are inputted
    order_offset = len(grid_1)
    grid = grid_1.copy()
    
    for order_no in grid_2:
        grid[order_no + order_offset] = grid_2[order_no]
    
    return grid
    
def last_level_crossed(data, base):
    ''' 
    Returns a list containing the last grid level touched.
    The grid levels are determined by the base input variable, 
    which corresponds to the pip_space x pip_value.
    '''
    # base = 20*0.0001
    
    last_level_crossed = np.nan
    levels_crossed = []
    for i in range(len(data)):
        high = data.High.values[i]
        low = data.Low.values[i]
        
        upper_prices = []
        lower_prices = []
        
        for price in [high, low]:    
            upper_prices.append(base*np.ceil(price/base))
            lower_prices.append(base*np.floor(price/base))
        
        if lower_prices[0] != lower_prices[1]:
            # Candle has crossed a level
            last_level_crossed = lower_prices[0]
        
        levels_crossed.append(last_level_crossed)
    
    return levels_crossed


def build_multiplier_grid(origin, direction, multiplier, no_levels, precision, spacing):
    '''
    Constructs grid levels with a multiplying grid space.
    
        Parameters:
            origin (float): origin of grid as price amount.
            
            direction (int): direction of grid (1 for long, -1 for short).
            
            multiplier (float): grid space multiplier when price moves away 
            from the origin opposite to direction.
            
            no_levels (int): number of levels to calculate either side of the 
            origin.
            
            precision (int): instrument precision (eg. 4 for most currencies, 2 
            for JPY).
            
            spacing (float): spacing of grid in price units.
    '''
    
    levels = [i for i in range(1, no_levels + 1)]

    pos_levels = [round(origin + direction*spacing*i, precision) for i in levels]
    neg_spaces = [spacing*multiplier**(i) for i in levels]
    neg_levels = []
    prev_neg_level = origin
    for i in range(len(levels)):
        next_neg_level = prev_neg_level - direction*neg_spaces[i]
        prev_neg_level = next_neg_level
        neg_levels.append(round(next_neg_level, precision))
    
    grid = neg_levels + [origin] + pos_levels
    grid.sort()
    
    return grid


def last_level_touched(data, grid):
    '''
    Calculates the grid levels touched by price data.
    '''
    
    # initialise with nan
    last_level_crossed = np.nan 
    
    levels_touched = []
    for i in range(len(data)):
        high = data.High.values[i]
        low = data.Low.values[i]
        
        upper_prices = []
        lower_prices = []
        
        for price in [high, low]:    
            # Calculate level above
            upper_prices.append(grid[next(x[0] for x in enumerate(grid) if x[1] > price)])
            
            # calculate level below
            first_level_below_index = next(x[0] for x in enumerate(grid[::-1]) if x[1] < price)
            lower_prices.append(grid[-(first_level_below_index+1)])
        
        if lower_prices[0] != lower_prices[1]:
            # Candle has crossed a level, since the level below the candle high
            # is different to the level below the candle low.
            # This essentially means the grid level is between candle low and high.
            last_level_crossed = lower_prices[0]
        
        levels_touched.append(last_level_crossed)
    
    return levels_touched

